//
// Author: Ahmed Ayadi (Ahmed.Ayadi@sophia.inria.fr)
// Generalization: Andras Varga, Emin Ilker Cetinbas
// Copyright (C) 2007 Ahmed Ayadi Emin Ilker Cetinbas, Andras Varga
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//


#ifndef MUSEUM_MOBILITY_H
#define MUSEUM_MOBILITY_H

#include <omnetpp.h>

#include "BasicMobility.h"


/**
 * @brief Models the mobility of with mass, making random motions.
 * See NED file for more info.
 *
 * @ingroup mobility
 * @author Ahmed Ayadi, Emin Ilker Cetinbas, Andras Varga
 */
class INET_API MuseumMobility : public BasicMobility
{
	enum 
    {
        WALKING,   // The station is walking
        PAUSE,  // The station in pause
    }State; 

  protected:
    // config (see NED file for explanation)
    cPar *changeAngleInterval;
    cPar *changeAngleBy;
    cPar *speed;
	cPar *pausePeriod;   ///< time of pause
    cPar *walkingPeriod;  ///< time of walking

    // current state
    double currentSpeed;   ///< speed of the host
    double currentAngle;   ///< angle of linear motion
	double upDatePositionInterval;

    Coord step;            ///< calculated from speed, angle and updateInterval
	double maxDistance;    ///< max distance from Access point 

	double nextUpdate;
	double lastUpdate;
	double walkingStart;

	double xMin;
	double xMax;
    double yMin;
	double yMax;

  protected:
    /** @name Timer messages */
    //@{
    cMessage *endMove;
    cMessage *endTurn;
    cMessage *endWalking;
    cMessage *endUpdate;
    //@}
  public:
    /**
     * @name Construction functions
     */
    //@{
    MuseumMobility();
    virtual ~MuseumMobility();
    //@}

  protected:
    /** @brief Initializes mobility model parameters.*/
    virtual void initialize(int);

   protected:
    /** @brief Called upon arrival of a self messages*/
    virtual void handleSelfMsg(cMessage *msg);

    /** @brief Move the host*/
    void move();
};

#endif