#include "HandleNotificationsClass.h"

HandleNotificationsClass::HandleNotificationsClass(void)
{

}

HandleNotificationsClass::~HandleNotificationsClass(void)
{

}


int HandleNotificationsClass::CheckNotificationMsg(cMessage *msg, MACAddress* destAddress, MACAddress* apAddress, int *numFrames, Ieee80211DataFrame *frameNotif, MACAddress myAddress)
{
	APNotification *apNotifMsg = NULL;
	EtherFrame* ethFrame = check_and_cast<EtherFrame *>(msg);
	EtherFrame* ethFrameToTreat = (EtherFrame *) ethFrame->dup();
	apNotifMsg = (APNotification *) ethFrameToTreat->decapsulate();
	
	int msgID = apNotifMsg->getMsgID();
	int status;

	switch(msgID)
	{
		case ASS_NOTIF:
		std::cout << "Node " << apNotifMsg->getSrcMACAddress() << " is associated with " << apNotifMsg->getAssAPMACAddress() << " at " << simTime() << "\n";
		status = AddOrModifyDestInfo(apNotifMsg->getSrcMACAddress(), apNotifMsg->getAssAPMACAddress());
		*destAddress = apNotifMsg->getSrcMACAddress();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		frameNotif = NULL;

		if (status == DEST_INFO_ENTRY_ADDED)
		{
			EV << "CheckNotificationMsg, Entry Added in the destination info list\n";
		}
		else if (status == DEST_INFO_ENTRY_MODIFIED)
		{
			EV << "CheckNotificationMsg, Entry found and modified in the destination info list\n";
		}
		else
		{
			EV << "CheckNotificationMsg, Error in entering destination info\n";
		}		
		break;

	case DISASS_NOTIF:
		std::cout << "Node " << apNotifMsg->getSrcMACAddress() << " is dis-associated with " << apNotifMsg->getAssAPMACAddress() << " at " << simTime() << "\n";
		*numFrames = apNotifMsg->getNumFramesRemoved();
		*destAddress = apNotifMsg->getSrcMACAddress();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		frameNotif = NULL;

		status = RemoveDestInfo(apNotifMsg->getSrcMACAddress());

		if (status == DEST_INFO_ENTRY_REMOVED)
		{
			EV << "CheckNotificationMsg, Entry removed from the destination info list\n";
		}
		else if (status == DEST_INFO_ENTRY_NOT_FOUND)
		{
			EV << "CheckNotificationMsg, Entry not found in the destination info list\n";
		}
		else
		{
			EV << "CheckNotificationMsg, Error in removing entry\n";
		}

		status = CheckAndAddDestBufferAP(apNotifMsg->getSrcMACAddress(), apNotifMsg->getAssAPMACAddress());
		break;

	case DEST_FOUND_NOTIF:
		break;

	case UPDATE_FRAMES_NOTIF:
		std::cout << "Message From AP " << apNotifMsg->getAssAPMACAddress() << ": Buffered Frames for Node " << apNotifMsg->getSrcMACAddress() << " needs to be updated \n";
		*numFrames = apNotifMsg->getNumFramesRemoved();
		*destAddress = apNotifMsg->getSrcMACAddress();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		frameNotif = NULL;
		break;

	case REMOVE_FRAMES_NOTIF:
		*destAddress = apNotifMsg->getSrcMACAddress();
		*numFrames = apNotifMsg->getNumFramesRemoved();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		frameNotif = NULL;
		std::cout << "CheckNotificationMsg, REMOVE_FRAMES_NOTIF arrived from " << apNotifMsg->getAssAPMACAddress() << " at " << simTime() << "\n";
		break;

	case STORE_FRAMES:
		std::cout << "CheckNotificationMsg, STORE_FRAMES arrived from " << apNotifMsg->getAssAPMACAddress() << " at " << myAddress << " for " << apNotifMsg->getSrcMACAddress() << " at " << simTime() << "\n";
		*destAddress = apNotifMsg->getSrcMACAddress();
		*numFrames = apNotifMsg->getNumFramesRemoved();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		*frameNotif = apNotifMsg->getFrame();
		break;

	case FETCH_FRAMES:
		std::cout << "CheckNotificationMsg, FETCH_FRAMES arrived from " << apNotifMsg->getAssAPMACAddress() << " for " << apNotifMsg->getSrcMACAddress() << " at " << simTime() << "\n";		
		*destAddress = apNotifMsg->getSrcMACAddress();
		*numFrames = apNotifMsg->getNumFramesRemoved();
		*apAddress = apNotifMsg->getAssAPMACAddress();
		*frameNotif = apNotifMsg->getFrame();
		break;

	default:
//		std::cout << "Message for Node" << apNotifMsg->getSrcMACAddress() << " is unknown for AP " << apNotifMsg->getAssAPMACAddress() << "\n";
		delete ethFrameToTreat;
		delete apNotifMsg;
		return FAILURE;
		break;
	}

	delete ethFrameToTreat;
	delete ethFrame;
	delete apNotifMsg;
	return msgID;
}


EtherFrame* HandleNotificationsClass::SendNotificationToAPs(int notifType, MACAddress nodeAddress, MACAddress apAddress, int numFrames, Ieee80211DataFrame* frame)
{
	APNotification *notifMsg = new APNotification;

	notifMsg->setMsgID(notifType);
	notifMsg->setSrcMACAddress(nodeAddress);
	notifMsg->setAssAPMACAddress(apAddress);
	notifMsg->setNumFramesRemoved(numFrames);
	if (frame != NULL)
		notifMsg->setFrame(*frame);
	delete frame;

	EtherFrame *ethframe = new EthernetIIFrame(); //TODO option to use EtherFrameWithSNAP instead//	ethframe->setDest(frame->getAddress3());
	ethframe->setSrc(apAddress);
	ethframe->setDest(MACAddress::BROADCAST_ADDRESS);
	ethframe->setKind(ETH_FRAME);
	cMessage *payload = (cMessage*) notifMsg;

	ASSERT(payload!=NULL);
	ethframe->encapsulate(payload);

	return ethframe;
}


int HandleNotificationsClass::CheckForDestInfo(MACAddress frameAddress, MACAddress* apAddress)
{	
	std::list<struct DestInfo>::iterator newIterator;		// link list iterator

	for (newIterator = destInfoList.begin(); newIterator != destInfoList.end(); newIterator++)
	{
		if ((newIterator)->destNodeAddress == frameAddress)
		{
			*apAddress = (newIterator)->assAPAddress;
			EV << "CheckForDestInfo, Search Dest Info, Entry found\n";
			return DEST_INFO_ENTRY_FOUND;
		}
	}

	EV << "CheckForDestInfo, Search Dest Info, Entry not found\n";
	return DEST_INFO_ENTRY_NOT_FOUND;
}


int HandleNotificationsClass::AddOrModifyDestInfo(MACAddress srcMACAddress, MACAddress assAPMACAddress)
{
	struct DestInfo tempDestInfo;

	std::list<struct DestInfo>::iterator newIterator;		// link list iterator
	
	for (newIterator = destInfoList.begin(); newIterator != destInfoList.end(); newIterator++)
	{
		if ((newIterator)->destNodeAddress == srcMACAddress)
		{
			(newIterator)->assAPAddress = assAPMACAddress;
			return DEST_INFO_ENTRY_MODIFIED;
		}
	}

	tempDestInfo.destNodeAddress = srcMACAddress;
	tempDestInfo.assAPAddress = assAPMACAddress;
	destInfoList.push_back(tempDestInfo);
	
	return DEST_INFO_ENTRY_ADDED;
}


int HandleNotificationsClass::RemoveDestInfo(MACAddress srcMACAddress)
{
	std::list<struct DestInfo>::iterator newIterator;		// link list iterator

	for (newIterator = destInfoList.begin(); newIterator != destInfoList.end(); newIterator++)
	{
		if ((newIterator)->destNodeAddress == srcMACAddress)
		{
			destInfoList.erase(newIterator);			
			return DEST_INFO_ENTRY_REMOVED;
		}
	}

	return DEST_INFO_ENTRY_NOT_FOUND;
}

int HandleNotificationsClass::CheckAndAddDestBufferAP(MACAddress srcMACAddress, MACAddress assAPMACAddress)
{
	struct DestBufferInfo tempDestBufferInfo;
	
	std::list<struct DestBufferInfo>::iterator newIterator;		// link list iterator
	
	for (newIterator = destBufferInfoList.begin(); newIterator != destBufferInfoList.end(); newIterator++)
	{
		if ((newIterator)->destNodeAddress == srcMACAddress)
		{
			newIterator->respAPAddress = assAPMACAddress;
			return DEST_BUFFER_ENTRY_MODIFIED;
		}
	}

	EV << "HandleNotificationsClass::CheckAndAddDestBufferAP(), adding buffering AP to destination = " << srcMACAddress << endl;
	tempDestBufferInfo.destNodeAddress = srcMACAddress;
	tempDestBufferInfo.respAPAddress = assAPMACAddress;
	destBufferInfoList.push_back(tempDestBufferInfo);	
	
	return DEST_BUFFER_AP_ADDED;
}

int HandleNotificationsClass::CheckForDestBufferInfo(MACAddress frameAddress, MACAddress* apAddress)
{	
	std::list<struct DestBufferInfo>::iterator newIterator;		// link list iterator

	for (newIterator = destBufferInfoList.begin(); newIterator != destBufferInfoList.end(); newIterator++)
	{
		if ((newIterator)->destNodeAddress == frameAddress)
		{
			*apAddress = (newIterator)->respAPAddress;
			EV << "CheckForDestBufferInfo, Search Dest Info, Entry found\n";			
			return DEST_BUFFER_INFO_ENTRY_FOUND;
		}
	}

	EV << "CheckForDestBufferInfo, Search Dest Info, Entry not found\n";
	return DEST_BUFFER_INFO_ENTRY_NOT_FOUND;
}