#include "buffer.h"


BufferClass::BufferClass()
{
	listSize_ = 0;
	listBufferLimit_ = 5000;
	lowPriorityLossCounter_ = 0;
	highPriorityLossCounter_ = 0;
	ttlExpiredCounter_ = 0;
}


void BufferClass::AddElementToList(bufferedFrames_t frameToAdd)
{
	bufferedFrames_t thisP;		// temporary variable to keep an entry of a buffered element

	// Check if the buffer has any space left to save this element
	if (listSize_ < listBufferLimit_)
	{// if space is available
		// save the element in the buffer
		// packet should be added at the back of the list
		ptrPktList_.push_back(frameToAdd);
		std::cout << "BufferClass::Adding Element to List " << listSize_ << endl;
		// add the size of the buffer by one
		listSize_++;
	}
	else if (IsPacketTagged(frameToAdd))
	{// if no space is available in the buffer and coming frame is of high priority
		std::list<bufferedFrames_t>::iterator newIterator;		// link list iterator
		bool removeFlag = false;	// flag to keep track if an entry is removed from the buffer

		// for loop to iterate through the buffer list
		// we don't increment iterator element in foor loop
		// this is because the iterator is automatically advanced by one if we remove an element
		for (newIterator = ptrPktList_.begin(); newIterator != ptrPktList_.end(); )
		{
			thisP = *newIterator;	// this iteration of the buffered element
			
			// check if this element is tagged (prioritized)
			if (!IsPacketTagged(thisP))
			{// if this element is of low priority
				// we can remove this element from the buffer to accomodate the newly arrived frame
				EV << "BufferClass::AddElementToList, Dropping Untagged Element\n";
				lowPriorityLossCounter_++;		// another low priority frame is dropped

				newIterator = ptrPktList_.erase(newIterator);	// erase oldest low priority element from the buffer
 				listSize_--;		// list size is decremented
				removeFlag = true;	// an entry is removed, set flag to true

				EV << "BufferClass::AddElementToList, Adding Packet after removing one\n";
				// add new element to buffer now
				ptrPktList_.push_back(frameToAdd); // packet should be added at the back of the list
				listSize_++;		// increment list size

				break;
			}
			else
			{// if current element is of high priority
				newIterator++;		// go to next iterator to see if it is of low priority
			}
		}
		
		if (!removeFlag)
		{// if no element is removed
			// buffer is still full and all elements are of high priority
			// Remove oldest entered high pririty element
			EV << "BufferClass::AddElementToList, Dropping Oldest Element\n";
			highPriorityLossCounter_++;			// another high priority frame is dropped

			std::list<bufferedFrames_t>::iterator newIterator;		// link list iterator
			newIterator = ptrPktList_.begin();	// go to the start of the list to remove one element
			newIterator = ptrPktList_.erase(newIterator);	// remove the oldest high priority element
			listSize_--;		// decrement list size by one
			
			// double check if the buffer has any space
			if (listSize_ < listBufferLimit_)
			{
				// add the new element to the end of the list
				EV << "BufferClass::AddElementToList, Adding Packet after removing two\n";
				ptrPktList_.push_back(frameToAdd); // packet should be added at the back of the list rather
				listSize_++;	// increment list size by one
			}			
		}
	}
	else
	{// new frame is of low priority and the buffer is full with high priority frames
		
		std::list<bufferedFrames_t>::iterator newIterator;		// link list iterator
		bool removeFlag = false;	// flag to keep track if an entry is removed from the buffer

		// for loop to iterate through the buffer list
		// we don't increment iterator element in foor loop
		// this is because the iterator is automatically advanced by one if we remove an element
		for (newIterator = ptrPktList_.begin(); newIterator != ptrPktList_.end(); )
		{
			thisP = *newIterator;	// this iteration of the buffered element
			
			// check if this element is tagged (prioritized)
			if (!IsPacketTagged(thisP))
			{// if this element is of low priority
				// we can remove this element from the buffer to accomodate the newly arrived frame
				EV << "BufferClass::AddElementToList, Dropping Untagged Element\n";
				lowPriorityLossCounter_++;		// another low priority frame is dropped

				newIterator = ptrPktList_.erase(newIterator);	// erase oldest low priority element from the buffer
 				listSize_--;		// list size is decremented
				removeFlag = true;	// an entry is removed, set flag to true

				EV << "BufferClass::AddElementToList, Adding Packet after removing one\n";
				// add new element to buffer now
				ptrPktList_.push_back(frameToAdd); // packet should be added at the back of the list
				listSize_++;		// increment list size

				break;
			}
			else
			{// if current element is of high priority
				newIterator++;		// go to next iterator to see if it is of low priority
			}
		}

		if (!removeFlag)
		{// all elements in the buffer are of high priority, we discard the newly arrived frame
			EV << "BufferClass::AddElementToList, Discarding the packet, Buffer is full\n";
			lowPriorityLossCounter_++;				// another low priority frame is dropped
		}
	}
}


int BufferClass::IsPacketTagged(bufferedFrames_t frameToBuffer)
{
	// check the priority field of the frame
	if (frameToBuffer.priority == 2)
	{
		// the frame is of high priority
		return true;
	}
	else
	{
		// the frame is of low priority
		return false;
	}
}

/*
void BufferClass::DecrementTTLValueOfElements(void)
{
	printf("BufferClass::DecrementTTLValueOfElements, Decrementing TTL for Packets\n");
	
	std::list<Ieee80211DataFrame*>::iterator newIterator;		// link list iterator
	Ieee80211DataFrame* thisP;	
/*
	hdr_udpbuffer* hp;
	hdr_ip* ip;

	for (newIterator = ptrPktList_.begin(); newIterator != ptrPktList_.end(); newIterator++)
	{
		thisP = (Ieee80211DataFrame*) *newIterator;		

		hp = hdr_udpbuffer::access(thisP);
		ip = hdr_ip::access(thisP);
		hp->ttl_--;
		
		if (hp->ttl_ == 0)
		{
			printf("BufferClass::DecrementTTLValueOfElements, TTL Expired, Dropping Packet\n");
//			CheckAndAddEntryInDropList(ip->daddr(), TTL_EXP_DROP);
			ttlExpiredCounter_++;
			newIterator = ptrPktList_.erase(newIterator);
			listSize_--;
		}
	}
*/
/*}

/*
void BufferClass::CheckAndAddEntryInDropList(nsaddr_t destAddress, int type)
{
//	list<dropPacketsInfo_t*>::iterator dpiIterator;		// link list iterator
	int isEntryExist = FALSE;
				
//	for (dpiIterator = ptrDropPacketsInfo_.begin(); dpiIterator != ptrDropPacketsInfo_.end(); dpiIterator++)
	for (int i=0; i<sizeDPI_; i++)
	{
		//if ((*dpiIterator)->nodeAddress_ == destAddress)
		if (dropPacketsInfo_[i].nodeAddress_ == destAddress)
		{
			isEntryExist = TRUE;
			switch (type)
			{
				case LP_PKT_DROP:
				{
					//((*dpiIterator)->lowPriorityLossCounter_)++;
					(dropPacketsInfo_[i].lowPriorityLossCounter_)++;
					break;
				}
				case HP_PKT_DROP:
				{
					//((*dpiIterator)->highPriorityLossCounter_)++;
					(dropPacketsInfo_[i].highPriorityLossCounter_)++;
					break;
				}
				case TTL_EXP_DROP:
				{
					//((*dpiIterator)->ttlExpiredCounter_)++;
					(dropPacketsInfo_[i].ttlExpiredCounter_)++;
					break;
				}
				default:
				{
					printf("Wrong TYPE of Entry\n");
					break;
				}
			}
		}
	}
	
	if (!isEntryExist)
	{
		//struct dropPacketsInfo_t* temp = new (struct dropPacketsInfo_t);

		// Set the fields of the structure to add the entry in the list
		//temp->nodeAddress_ = destAddress;	// Set the mobile node address
		
		dropPacketsInfo_[sizeDPI_].nodeAddress_ = destAddress;
		
		switch (type)
		{
			case LP_PKT_DROP:
			{
				/*
				temp->lowPriorityLossCounter_ = 1;	// Set the counter to 1
				temp->highPriorityLossCounter_ = 0;	// Set the counter to 0
				temp->ttlExpiredCounter_ = 0;		// Set the counter to 0
				*/
/*				dropPacketsInfo_[sizeDPI_].lowPriorityLossCounter_ = 1;
				dropPacketsInfo_[sizeDPI_].highPriorityLossCounter_ = 0;
				dropPacketsInfo_[sizeDPI_].ttlExpiredCounter_ = 0;
				break;
			}
			case HP_PKT_DROP:
			{
				/*
				temp->lowPriorityLossCounter_ = 0;	// Set the counter to 1
				temp->highPriorityLossCounter_ = 1;	// Set the counter to 0
				temp->ttlExpiredCounter_ = 0;		// Set the counter to 0
				*/
/*				dropPacketsInfo_[sizeDPI_].lowPriorityLossCounter_ = 0;
				dropPacketsInfo_[sizeDPI_].highPriorityLossCounter_ = 1;
				dropPacketsInfo_[sizeDPI_].ttlExpiredCounter_ = 0;
				break;
			}
			case TTL_EXP_DROP:
			{
				/*
				temp->lowPriorityLossCounter_ = 0;	// Set the counter to 1
				temp->highPriorityLossCounter_ = 0;	// Set the counter to 0
				temp->ttlExpiredCounter_ = 1;		// Set the counter to 0
				*/
/*				dropPacketsInfo_[sizeDPI_].lowPriorityLossCounter_ = 0;
				dropPacketsInfo_[sizeDPI_].highPriorityLossCounter_ = 0;
				dropPacketsInfo_[sizeDPI_].ttlExpiredCounter_ = 1;
				break;
			}
			default:
			{
				printf("Wrong TYPE of Entry\n");
				break;
			}
		}
		
		//ptrDropPacketsInfo_.push_front(temp);
		sizeDPI_++;
		//delete temp;
	}
}
*/


