//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Authors:   Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
//           Masood Khosroshahy < m.khosroshahy@iee.org>
//
#include "FreeSpaceModel.h"
#include "FWMath.h"

Register_Class(FreeSpaceModel);

void FreeSpaceModel::initializeFrom(cModule *radioModule)
{
    EV<< "FreeSpaceModel selected "<< endl;
    systemLoss = radioModule->par("systemLoss");
    receiverAntennaGain = radioModule->par("receiverAntennaGain");
}

double FreeSpaceModel::calculateReceivedPower(double pSend, double carrierFrequency, double distance)
{
    const double speedOfLight = 300000000.0;
    double power;
    double waveLength = speedOfLight / carrierFrequency;
    EV<<"ReceivedPower"<< pSend * receiverAntennaGain * waveLength * waveLength / (16 * M_PI * M_PI * pow(distance,2) * systemLoss)<<endl;
    power=pSend * receiverAntennaGain * waveLength * waveLength / (16 * M_PI * M_PI * pow(distance,2) * systemLoss);
    return (power);
}

