//
// Copyright (c) 2007 INRIA
// All rights reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Authors:   Ahmed Ayadi < Ahmed.Ayadi@sophia.inria.fr>
//           Masood Khosroshahy < m.khosroshahy@iee.org>
//
#ifndef SHADOWINGMODEL_H
#define SHADOWINGMODEL_H

#include <itpp/itbase.h>
#include <itpp/itcomm.h>

#include "IReceptionModel.h"
#include "FWMath.h"

/**
 * Shadowing large-scale path loss model which calculates the received power by calculating
 * the received power at a reference d0 (set to 1 here), then incorporating the effect
 * of path loss exponent and shadowing variance.
 */
class INET_API ShadowingModel : public IReceptionModel
{
  private:
    double systemLoss;
    double receiverAntennaGain;
    double shadowing;
    double pathLossExponent;
    double shadowingVariance;
    int shadowingNumberofSamples;
    int shadowingRandomNumberVectorIndex;
    
    /** @brief The vector to store the generated random numbers */
    itpp::vec shadowingRandomNumberVector; 
    
  public:

    virtual void initializeFrom(cModule *radioModule);

    /**
     * Perform the calculation.
     */
    virtual double calculateReceivedPower(double pSend, double carrierFrequency, double distance);
};

#endif

