/***************************************************************************
                          RTPAVProfilePayload32Receiver.cc  -  description
                             -------------------
    begin                : Sun Jan 6 2002
    copyright            : (C) 2002 by Matthias Oppitz
    email                : Matthias.Oppitz@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/** \file RTPAVProfilePayload32Receiver.cc
 *  * In this file member functions of RTPAVProfilePayload32Receiver are
 *  * implemented.
 *  */

#include <omnetpp.h>

#include "RTPPayloadReceiver.h"
#include "RTPAVProfilePayload32Receiver.h"
#include "RTPPacket.h"
#include "RTPMpegPacket.h"

Define_Module_Like(RTPAVProfilePayload32Receiver, RTPPayloadReceiver);


RTPAVProfilePayload32Receiver::~RTPAVProfilePayload32Receiver() {
    delete _queue;
    delete packetLostVector;
    delete receivedThroughput;
};


void RTPAVProfilePayload32Receiver::initialize() {
    RTPPayloadReceiver::initialize();
    _payloadType = 32;
    _queue = new cQueue("IncomingQueue", &(RTPPacket::compareFunction));
    _lowestAllowedTimeStamp = 0;
    _highestSequenceNumber = 0;
    _lostPacketNumber = 0;
    _lowestSequenceNumber = 0;
    _packetLossRate = 0;
    timeStampLastMessageReceived = 0;
    samplingCoeff = 100;
    packetLostVector = new cOutVector("RTP Loss Rate");
    receivedThroughput = new cOutVector("Goodput RTP");
    WATCH(_lowestAllowedTimeStamp);
    WATCH(_highestSequenceNumber);
    WATCH(_lostPacketNumber);
    WATCH(_lowestSequenceNumber);
    WATCH(_packetLossRate);
};


void RTPAVProfilePayload32Receiver::processPacket(RTPPacket *rtpPacket)
{
    if(i%samplingCoeff==0)
    {
        receivedThroughput->record(recvdThroughput);
        i=0;
        recvdThroughput=0;
    }
    i++;

    if (timeStampLastMessageReceived == 0)
        timeStampLastMessageReceived = simTime();
    else
    {
        recvdThroughput+=((rtpPacket->length()/(simTime()-timeStampLastMessageReceived))/1000000)/samplingCoeff;
        timeStampLastMessageReceived = simTime();
    }

    // the first packet sets the lowest allowed time stamp
    if (_lowestAllowedTimeStamp == 0)
    {
        _lowestAllowedTimeStamp = rtpPacket->timeStamp();
        _lowestSequenceNumber = rtpPacket->sequenceNumber();
        _highestSequenceNumber = rtpPacket->sequenceNumber();
        _outputLogLoss <<"sequenceNumberBase"<< rtpPacket->sequenceNumber() << endl;
    }
    else
    {
        if (rtpPacket->sequenceNumber() > _highestSequenceNumber+1)
        {
            for (int i = _highestSequenceNumber+1; i< rtpPacket->sequenceNumber(); i++ )
            {
                _outputLogLoss << i << endl;
            }
            _lostPacketNumber = rtpPacket->sequenceNumber() - _highestSequenceNumber;
            _packetLossRate = (_lostPacketNumber / ((double)rtpPacket->sequenceNumber() - (double)_lowestSequenceNumber)) /2;
//             std::cout<<parentModule()->fullPath()<<"."<<className()<<":"<<"PacketLost : "<< _lostPacketNumber<< "  "<<rtpPacket->sequenceNumber() - _lowestSequenceNumber<<" PacketLossRate= "<<_packetLossRate<<endl;
        }
    }

    if ((rtpPacket->sequenceNumber()-_lowestSequenceNumber)%20 == 0)
    {
        RTPInnerPacket *rinpOut = new RTPInnerPacket("ReceiverStats");
        rinpOut->receiverModuleStats(_packetLossRate);
        packetLostVector->record(_packetLossRate);
        send(rinpOut, "toProfile");
    }

    if ((rtpPacket->timeStamp() < _lowestAllowedTimeStamp) || (rtpPacket->sequenceNumber()<= _highestSequenceNumber)) {
        delete rtpPacket;
    }
    else {
        // is this packet from the next frame ?
        // this can happen when the marked packet has been
        // lost or arrives late
        bool nextTimeStamp = rtpPacket->timeStamp() > _lowestAllowedTimeStamp;

        // is this packet marked, which means that it's
        // the last packet of this frame
        bool marked = rtpPacket->marker();

        // test if end of frame reached

        // check if we received the last (= marked)
        // packet of the frame or
        // we received a packet of the next frame
        _highestSequenceNumber = rtpPacket->sequenceNumber();

        if (nextTimeStamp || marked) {

            // a marked packet belongs to this frame
            if (marked && !nextTimeStamp) {
                _queue->insert(rtpPacket);
            }

            int pictureType = 0;
            int frameSize = 0;

            // the queue contains all packets for this frame
            // we have received
            while (!_queue->empty()) {
                RTPPacket *readPacket = (RTPPacket *)(_queue->head());
                _queue->remove(readPacket);
                RTPMpegPacket *mpegPacket = (RTPMpegPacket *)(readPacket->decapsulate());
                if (pictureType == 0)
                    pictureType = mpegPacket->pictureType();
                frameSize = frameSize + mpegPacket->payloadLength();

                delete mpegPacket;
                delete readPacket;
            };

            // we start the next frame
            // set the allowed time stamp
            if (nextTimeStamp) {
                _lowestAllowedTimeStamp = rtpPacket->timeStamp();
                _queue->insert(rtpPacket);
            };

            // we have calculated a frame
            if (frameSize > 0) {
                char line[100];
                // what picture type is it
                char picture = ' ';
                if (pictureType == 1)
                    picture = 'I';
                else if (pictureType == 2)
                    picture = 'P';
                else if (pictureType == 3)
                    picture = 'B';
                else if (pictureType == 4)
                    picture = 'D';

                // create sim line
                sprintf(line, "%f %i %c-Frame", simTime(), frameSize * 8, picture);
                // and write it to the file
                _outputFileStream << line << endl; 
            }
        }
        // we are not at the end of the frame
        // so just insert this packet
        else {
            _queue->insert(rtpPacket);
        }
    }
};
